const std = @import("std");
const mem = std.mem;
const heap = std.heap;

pub const c = @cImport({
    @cInclude("libvirt/libvirt.h");
    @cInclude("libvirt/libvirt-admin.h");
    @cInclude("libvirt/libvirt-lxc.h");
    @cInclude("libvirt/libvirt-qemu.h");
    @cInclude("libvirt/virterror.h");
});

fn handleError() VirError {
    const err = c.virGetLastError();
    std.debug.print("err: {any}\n", .{err});
    return VirError.OK;
    // switch (err.*.code) {
    //     0 => return VirError.OK,
    //     1 => return VirError.InternalError,
    //     2 => return VirError.NoMemory,
    //     // TODO rest
    //     else => VirError.OK,
    // }
}

pub const VirError = error{
    OK,
    InternalError,
    NoMemory,
    NoSupport,
    UnknownHost,
    NoConnect,
    InvalidConn,
    InvalidDomain,
    InvalidArg,
    OperationFailed,
    GetFailed,
    PostFailed,
    HttpError,
    SExprError,
    NoXen,
    XenCall,
    OsType,
    NoKernel,
    NoRoot,
    NoSource,
    NoTarget,
    NoName,
    NoOs,
    NoDevice,
    NoXenstore,
    DriverFull,
    CallFailed,
    XmlError,
    DomExist,
    OperationDenied,
    OpenFailed,
    ReadFailed,
    ParseFailed,
    ConfSyntax,
    // TODO rest
};

fn string(str: [*c]const u8) []const u8 {
    return mem.span(str);
}

fn intFromFlags(comptime T: type, flags: []const T) c_uint {
    var flags_int: c_uint = 0;
    for (flags) |f| flags_int |= @intFromEnum(f);
    return flags_int;
}

fn Iterator(comptime T: type, comptime Ptr: type, comptime freeFn: *const fn (Ptr) callconv(.C) c_int) type {
    return struct {
        list: [*]Ptr,
        num: c_int,
        curr: usize,

        pub fn init(
            comptime PP: type, // Parent Pointer
            comptime P: type, // Parent
            comptime F: type, // Flags
            parent: *const P,
            flags: []const F,
            initFn: *const fn (?*PP, [*c][*c]Ptr, c_uint) callconv(.C) c_int,
        ) !Self {
            var list: [*]Ptr = undefined;
            const num = initFn(parent.ptr, @ptrCast(&list), intFromFlags(F, flags));
            return if (num < 0) handleError() else .{
                .list = list,
                .num = num,
                .curr = 0,
            };
        }
        pub fn deinit(self: *Self) void {
            var i: usize = 0;
            while (i < self.num) : (i += 1) _ = freeFn(self.list[i]);
        }
        pub fn first(self: *Self) T {
            self.curr = 0;
            return .{ .ptr = self.list[self.curr] };
        }
        pub fn next(self: *Self) ?T {
            if (self.curr >= self.num) return null;
            const ptr = self.list[self.curr];
            self.curr += 1;
            return .{ .ptr = ptr };
        }
        const Self = @This();
    };
}

pub const Connection = struct {
    ptr: c.virConnectPtr,
    allocator: mem.Allocator,

    pub fn connect(uri: []const u8, allocator: mem.Allocator) VirError!Connection {
        const connection = c.virConnectOpenAuth(@ptrCast(uri), c.virConnectAuthPtrDefault, 0);
        if (connection) |conn| return .{
            .ptr = conn,
            .allocator = allocator,
        } else return handleError();
    }

    pub fn close(self: *const Connection) void {
        _ = c.virConnectClose(self.ptr);
    }

    pub fn getURI(self: *const Connection) ![]u8 {
        const uri = c.virConnectGetURI(self.ptr);
        defer std.c.free(uri);
        const str = string(uri);
        return if (str.len == 0) handleError() else try self.allocator.dupe(u8, str);
    }

    pub fn freeURI(self: *const Connection, uri: []u8) void {
        self.allocator.free(uri);
    }

    pub fn numOfPools(self: *const Connection) !u32 {
        const num = c.virConnectNumOfStoragePools(self.ptr);
        return if (num < 0) handleError() else @intCast(num);
    }
    pub fn numOfDefinedPools(self: *const Connection) !u32 {
        const num = c.virConnectNumOfDefinedStoragePools(self.ptr);
        return if (num < 0) handleError() else @intCast(num);
    }

    pub fn iteratePools(self: *const Connection, flags: []const Pool.ListFlags) !Pool.PoolIterator {
        return Pool.PoolIterator.init(
            c.virConnect,
            Connection,
            Pool.ListFlags,
            self,
            flags,
            c.virConnectListAllStoragePools,
        );
    }

    pub fn numOfDomains(self: *const Connection) !u32 {
        const num = c.virConnectNumOfDomains(self.ptr);
        return if (num < 0) handleError() else @intCast(num);
    }

    pub fn numOfDefinedDomains(self: *const Connection) !u32 {
        const num = c.virConnectNumOfDefinedDomains(self.ptr);
        return if (num < 0) handleError() else @intCast(num);
    }

    pub fn iterateDomains(self: *const Connection, flags: []const Domain.ListFlags) !Domain.DomainIterator {
        return Domain.DomainIterator.init(
            c.virConnect,
            Connection,
            Domain.ListFlags,
            self,
            flags,
            c.virConnectListAllDomains,
        );
    }
};

pub const Pool = struct {
    ptr: c.virStoragePoolPtr,
    arena: heap.ArenaAllocator,

    pub fn init(ptr: c.virStoragePoolPtr, allocator: mem.Allocator) Pool {
        return .{
            .ptr = ptr,
            .arena = heap.ArenaAllocator.init(allocator),
        };
    }
    pub fn deinit(self: *const Pool) void {
        self.arena.deinit();
    }

    pub fn createVolume(pool: *const Pool, xml: []const u8, flags: []Volume.Flags) !Volume {
        const volume = c.virStorageVolCreateXML(pool, xml, intFromFlags(Volume.CreateFlags, flags));
        return .{ .ptr = volume };
    }

    pub fn iterateVolumes(self: *const Pool, flags: []const ListFlags) Volume.VolumeIterator {
        return Volume.VolumeIterator.init(
            c.virStorageVolumePtr,
            Pool,
            Volume.ListFlags,
            self,
            flags,
            c.virStoragePoolListAllVolumes,
        );
    }

    pub const ListFlags = enum(c_uint) {
        Inactive = 1 << 0,
        Active = 1 << 1,
        Persistent = 1 << 2,
        Transient = 1 << 3,
        Autostart = 1 << 4,
        NoAutostart = 1 << 5,
        Dir = 1 << 6,
        Fs = 1 << 7,
        NetFs = 1 << 8,
        Logical = 1 << 9,
        Disk = 1 << 10,
        Iscsi = 1 << 11,
        Scsi = 1 << 12,
        Mpath = 1 << 13,
        Rbd = 1 << 14,
        Sheepdog = 1 << 15,
        Gluster = 1 << 16,
        Zfs = 1 << 17,
        Vstorage = 1 << 18,
        IscsiDirect = 1 << 19,
    };

    pub const PoolIterator = Iterator(Pool, c.virStoragePoolPtr, c.virStoragePoolFree);

    pub const Volume = struct {
        ptr: c.virStorageVolPtr,

        pub fn getName(self: *const Volume) ![]const u8 {
            const name = c.virStorageVolGetName(self.ptr);
            defer std.c.free(name);
            const str = string(name);
            return if (str.len == 0) handleError() else try self.arena.allocator().dupe(u8, str);
        }

        pub const ListFlags = enum(c_uint) {};
        pub const CreateFlags = enum(c_uint) {
            PreallocMetadata = 1 << 0,
            Reflink = 1 << 1,
            Validate = 1 << 2,
        };

        pub const VolumeIterator = Iterator(Volume, c.virStorageVolPtr, c.virStorageVolFree);
    };
};

pub const Domain = struct {
    ptr: c.virDomainPtr,

    pub fn getName(self: *const Domain) []const u8 {
        const name = c.virDomainGetName(self.ptr);
        return mem.span(name);
    }

    pub fn isActive(self: *const Domain) bool {
        const active = c.virDomainIsActive(self.ptr);
        return if (active == 0) false else true;
    }

    pub const ListFlags = enum(c_uint) {
        Active = c.VIR_CONNECT_LIST_DOMAINS_ACTIVE,
        Inactive = c.VIR_CONNECT_LIST_DOMAINS_INACTIVE,
    };

    pub const DomainIterator = Iterator(Domain, c.virDomainPtr, c.virDomainFree);
};
